/*
 * ping.c - send ICMP Echo request packets
 *
 * Copyright (c) 2000, 2001 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ping.c,v 1.9 2001/01/05 14:06:44 fyre Stab $
 */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>

static char *rcsid = "$Id: ping.c,v 1.9 2001/01/05 14:06:44 fyre Stab $";

#define MY_EXIT_SUCCESS	0
#define MY_EXIT_NORESP	1
#define MY_EXIT_FAILURE	2

static void fatal (const char *fmt, ...);
static int in_cksum (unsigned short *buf, int len);
static void make_addr (const char *host, struct sockaddr_in *sa);
static void ms_to_tv (int ms, struct timeval *tv);
static void send_echo (struct sockaddr_in *dest);
static int tv_to_ms (const struct timeval *tv);
static void usage (void);
static int wait_response (struct sockaddr_in *src);

static int rawfd = -1;
static int timeout = 20;

int main (int argc, char *argv[])
{
	struct sockaddr_in to;
	const char *host;
	int fd;

	/*
	 * Temporarily drop our suid root privileges until we need them.
	 */
	if (seteuid (getuid ()) < 0)
		fatal ("seteuid: %s", strerror (errno));

	/*
	 * Ensure that the raw socket we're about to open will not be
	 * file descriptor 0, 1 or 2 for security reasons. Some operating
	 * systems will do this for us but it doesn't hurt to make sure.
	 *
	 * TODO: Remove this somehow.
	 */
	do {
		if ((fd = open ("/dev/null", O_RDWR)) < 0)
			fatal ("/dev/null: %s", strerror (errno));
	} while (fd <= 2);
	if (close (fd) < 0)
		fatal ("close: %s", strerror (errno));

	/*
	 * ping needs root privileges to open the raw ICMP socket. Once the
	 * socket is open, we permanently drop our privileges.
	 */
	if (seteuid (0) < 0)
		fatal ("seteuid: %s", strerror (errno));
	if ((rawfd = socket (PF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0)
		fatal ("socket: %s", strerror (errno));
	if (rawfd <= 2) exit (EXIT_FAILURE); /* should not happen */
	if (setuid (getuid ()) < 0)
		fatal ("setuid: %s", strerror (errno));

	/*
	 * We are now running as the calling user and have opened the raw
	 * ICMP socket.
	 */

	if ((argc != 2) && (argc != 3))
		usage ();
	host = argv[1];
	if (argc > 2) {
		/*
		 * Be fascist about the range of timeout values we accept.
		 * Timeouts of more than an hour are not at all useful
		 * in practice.
		 */
		timeout = atoi (argv[2]);
		if (timeout <= 0)
			fatal ("illegal timeout value");
		else if (timeout > 60 * 60)
			fatal ("illegal timeout value");
	}

	/*
	 * Ensure the hostname is of a reasonable length. We terminate rather
	 * than truncating because truncating a hostname may cause it to point
	 * to a different machine than intended.
	 */
	if (strlen (host) >= MAXHOSTNAMELEN)
		fatal ("host too long");

	make_addr (host, &to);

	send_echo (&to);
	if (wait_response (&to) != 0) {
		printf ("%s did not respond\n", host);
		exit (MY_EXIT_NORESP);
	}

	printf ("%s is alive\n", host);

	return MY_EXIT_SUCCESS;
}

/*
 * Send an ICMP_ECHO request to `dest'.
 */
static void send_echo (struct sockaddr_in *dest)
{
	unsigned char outpacket[sizeof (struct icmphdr)];
	struct icmphdr *ich;
	int nw;

	ich = (struct icmphdr *) outpacket;
	ich->type = ICMP_ECHO;
	ich->code = 0;
	ich->checksum = 0;
	ich->un.echo.sequence = 0;
	ich->un.echo.id = getpid() & 0xFFFF;
	ich->checksum = in_cksum ((u_short *) outpacket, sizeof (outpacket));

	if ((nw = sendto (rawfd, outpacket, sizeof (outpacket), 0, dest, sizeof (*dest))) < 0)
		fatal ("sendto: %s", strerror (errno));

	if (nw != sizeof (outpacket))
		fatal ("short write");
}

/*
 * Convert a host string (hostname or IP address) to a sockaddr_in structure.
 */
static void make_addr (const char *host, struct sockaddr_in *sa)
{
	struct hostent *he;
	unsigned short last;
	
	sa->sin_port = 0;
	sa->sin_family = AF_INET;

	if ((he = gethostbyname (host)) != NULL) {
		if (he->h_addrtype != AF_INET)
			fatal ("host is not an INET address");
		if (he->h_addr_list[0] == NULL)
			fatal ("host has no addresses");
		memcpy (&sa->sin_addr, he->h_addr_list[0], sizeof (sa->sin_addr));
		if (he->h_addr_list[1] != NULL)
			fprintf (stderr, "ping: warning: host has multiple addresses, using first\n");
	} else if (inet_aton (host, &sa->sin_addr) == 0)
		fatal ("unknown host");

	last = ntohl (sa->sin_addr.s_addr) & 0xFF;
	if ((last == 0x00) || (last == 0xFF))
		fatal ("cannot ping broadcast address");
}

/*
 * Convert number of milliseconds to timeval.
 */
static void ms_to_tv (int ms, struct timeval *tv)
{
	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}

/*
 * Convert timeval to number of milliseconds.
 */
static int tv_to_ms (const struct timeval *tv)
{
	return tv->tv_sec * 1000 + tv->tv_usec / 1000;
}

/*
 * Wait for an ICMP_ECHOREPLY from the host.
 */
static int wait_response (struct sockaddr_in *src)
{
	struct sockaddr_in from;
	socklen_t fromlen = sizeof (from);
	struct iphdr *iph;
	struct icmphdr *ich;
	unsigned char inpacket[IP_MAXPACKET];
	fd_set fdset;
	int toremain = timeout * 1000;
	struct timeval totval, before, after;
	int nrecv;

	/*
	 * Keep reading packets until we find the response to the packet
	 * we sent or time out.
	 */
	do {
		FD_ZERO (&fdset);
		FD_SET (rawfd, &fdset);
		
		if (gettimeofday (&before, NULL) < 0)
			fatal ("gettimeofday: %s", strerror (errno));
		ms_to_tv (toremain, &totval);
		if (select (rawfd + 1, &fdset, NULL, NULL, &totval) < 0)
			fatal ("select: %s", strerror (errno));
		if (gettimeofday (&after, NULL) < 0)
			fatal ("gettimeofday: %s", strerror (errno));
		toremain -= tv_to_ms (&after) - tv_to_ms (&before);
		if (toremain < 0)
			return 1;
		if (!FD_ISSET (rawfd, &fdset))
			return 1;
		if ((nrecv = recvfrom (rawfd, inpacket, sizeof (inpacket), 0, &from, &fromlen)) < 0)
			fatal ("recvfrom: %s", strerror (errno));
		/*
		 * Verify some assumptions we make about the kernel ICMP
		 * socket handling.
		 */
		if ((size_t) nrecv < sizeof (struct iphdr))
			fatal ("OOPS: packet too short");
		iph = (struct iphdr *) inpacket;
		if (iph->version != 4)
			fatal ("OOPS: wrong IP version");
		if (iph->protocol != IPPROTO_ICMP)
			fatal ("OOPS: packet is not ICMP");
		if (iph->ihl * 4 + sizeof (struct icmphdr) > (size_t) nrecv)
			fatal ("OOPS: packet too short");
		ich = (struct icmphdr *) (inpacket + iph->ihl * 4);
	} while ((from.sin_addr.s_addr != src->sin_addr.s_addr) || 
		(ich->type != ICMP_ECHOREPLY) || 
		(ich->un.echo.id != (getpid() & 0xFFFF)) || 
		(ich->un.echo.sequence != 0));
	
	return 0;
}

/*
 * Calculate IP checksum.
 */
static int in_cksum (unsigned short *buf, int len)
{
	int sum = 0;

	while (len > 1) {
		sum += *buf++;
		len -= 2;
	}
	
	if (len == 1) {
		union {
			unsigned short s;
			unsigned char c[2];
		} u;
		u.c[0] = *(unsigned char *) buf;
		u.c[1] = 0;
		sum += u.s;
	}

	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += (sum >> 16);
	return ~sum;
}

static void usage (void)
{
	fprintf (stderr, "usage: ping host [timeout]\n");
	exit (MY_EXIT_FAILURE);
}

/*
 * Display fatal error message and terminate.
 */
static void fatal (const char *fmt, ...)
{
	va_list ap;

	fputs ("ping: ", stderr);
	va_start (ap, fmt);
	vfprintf (stderr, fmt, ap);
	va_end (ap);
	fputc ('\n', stderr);

	exit (MY_EXIT_FAILURE);
}
